from typing import List, Union
from datetime import datetime, timedelta
from sqlalchemy.orm import Session

from girok.server.src.database import get_db
import girok.server.src.task.schemas as schemas
import girok.server.src.utils as general_utils
import girok.server.src.task.service as service
import girok.server.src.category.service as category_service
import girok.server.src.task.exceptions as exceptions
import girok.server.src.task.enums as task_enums

# @router.post(
#     "/",
#     status_code=status.HTTP_201_CREATED,
#     response_model=schemas.TaskCreateOut
# )
def create_task(task: schemas.TaskCreateIn):
    db = next(get_db())
    task['deadline'] = datetime.strptime(task['deadline'], '%Y-%m-%d %H:%M:%S')
    try:
        new_task = service.create_task(db, task)
        return {"success": True, "new_task": new_task}
    except Exception as e:
        return {"success": False, "detail": e.detail}
    

# @router.get(
#     "/",
#     status_code=status.HTTP_200_OK
#     # response_model=schemas.TaskOut
# )
def get_tasks(data):
    data = schemas.TaskGetIn(**data).dict()
    category = data['category']
    start_date = data['start_date']
    end_date = data['end_date']
    priority = data['priority']
    no_priority = data['no_priority']
    tag = data['tag']
    view = data['view']
    
    db = next(get_db())
    try:
        # Check start_date <= end_date
        start_date = datetime.strptime(start_date, '%Y-%m-%d %H:%M:%S')
        end_date = datetime.strptime(end_date, '%Y-%m-%d %H:%M:%S')
        if start_date > end_date:
            raise exceptions.InvalidDateWindowException(start_date, end_date)
        
        if category is None: # ALL tasks regardless of category
            cat_ids = category_service.get_subcategory_ids_by_parent_id(db, None) # top most categories
            cat_ids += [None]
        elif category == ['']: # Only "None category" category 
            cat_ids = [None]
        else: # Specified category
            cat_id, _ = category_service.get_last_cat_id(db, category)
            cat_ids = [cat_id]

        if view == task_enums.TaskView.category:
            tasks = service.get_tasks_by_category(
                db=db,
                cat_ids=cat_ids,
                start_date=start_date,
                end_date=end_date,
                priority=priority,
                no_priority=no_priority,
                tag=tag
            )
        elif view == task_enums.TaskView.list:
            tasks = service.get_tasks_as_list(
                db=db,
                cat_ids=cat_ids,
                start_date=start_date,
                end_date=end_date,
                priority=priority,
                no_priority=no_priority,
                tag=tag
            )
        else:
            raise Exception("Invalid task view!")
        return {"success": True, "tasks": tasks}
    except Exception as e:
        print(e)
        return {"success": False, "detail": e.detail}
    
    
# @router.delete(
#     "/{task_id}",
#     status_code=status.HTTP_204_NO_CONTENT
# )
def delete_task(
    task_id: int
):
    db = next(get_db())
    try:
        service.delete_task(db, task_id)
        return {"success": True}
    except Exception as e:
        return {"success": False, "detail": e.detail}
    
    
# @router.patch(
#     '/{task_id}/tag',
#     status_code=status.HTTP_200_OK,
# )
def change_task_tag(
    task_id: int,
    tag: schemas.ChangeTaskTagIn
):
    db = next(get_db())
    try:
        new_tag_name = tag['new_tag_name']
        updated_task = service.change_task_tag(db, task_id, new_tag_name)
        return {"success": True, "updated_task": updated_task}
    except Exception as e:
        return {"success": False, "detail": e.detail}


# @router.patch(
#     '/{task_id}/priority',
#     status_code=status.HTTP_200_OK,
# )
def change_task_priority(
    task_id: int,
    priority: schemas.ChangeTaskPriorityIn
):
    db = next(get_db())
    try:
        new_priority = priority['new_priority']
        updated_task = service.change_task_priority(db, task_id, new_priority)
        return {"success": True, "updated_task": updated_task}
    except Exception as e:
        return {"success": False, "detail": e.detail}


# @router.patch(
#     '/{task_id}/date',
#     status_code=status.HTTP_200_OK
# )
def change_task_date(
    task_id: int,
    data: schemas.ChangeTaskDateIn
):
    db = next(get_db())
    try:
        new_date = data['new_date']
        new_date = datetime.strptime(new_date, "%Y-%m-%d %H:%M:%S")
        updated_task = service.change_task_date(db, task_id, new_date)
        return {"success": True, "updated_task": updated_task}
    except Exception as e:
        print(e)
        return {"success": False, "detail": e.detail}


# @router.patch(
#     '/{task_id}/name',
#     status_code=status.HTTP_200_OK
# )
def change_task_name(
    task_id: int,
    data: schemas.ChangeTaskNameIn
):
    db = next(get_db())
    try:
        new_name = data['new_name']
        updated_task = service.change_task_name(db, task_id, new_name)
        return {"success": True, "updated_task": updated_task}
    except Exception as e:
        return {"success": False, "detail": e.detail}


# @router.get(
#     '/tags',
#     status_code=status.HTTP_200_OK,
#     response_model=schemas.TagOut
# )
def get_tags():
    db = next(get_db())
    try:
        tags = service.get_tags(db)
        return {"success": True, "tags": tags}
    except Exception as e:
        return {"success": False, "detail": e.detail}


# @router.get(
#     "/{task_id}",
#     status_code=status.HTTP_200_OK,
#     response_model=schemas.GetSingleTaskOut
# )
def get_single_task(task_id: int):
    db = next(get_db())
    try:
        task = service.get_single_task(db, task_id)
        return {"success": True, "task": task}
    except Exception as e:
        return {"success": False, "detail": e.detail}
    



from collections import defaultdict
from sqlalchemy.orm import Session
from sqlalchemy import and_

import girok.server.src.category.exceptions as exceptions
import girok.server.src.category.models as models
import girok.server.src.utils as global_utils

def create_category(db: Session, cat_data):
    new_cat = models.TaskCategory(**cat_data)
    db.add(new_cat)
    db.commit()
    db.refresh(new_cat)
    return global_utils.sql_obj_to_dict(new_cat)


def delete_category(db: Session, cat_id: int):
    cat = db.query(models.TaskCategory).\
        filter(
            and_(
               models.TaskCategory.task_category_id == cat_id 
            )).first()
    db.delete(cat)
    db.commit()
    return cat


def rename_category(db: Session, cat_id: int, new_name: str):
    cat = db.query(models.TaskCategory).\
        filter(
            and_(
                models.TaskCategory.task_category_id == cat_id
            )
            ).first()
    setattr(cat, "name", new_name)
    db.commit()
    
    
def move_category(db: Session, cat_id: int, new_pid: int):
    cat = db.query(models.TaskCategory).\
        filter(
            and_(
                models.TaskCategory.task_category_id == cat_id
            )
            ).first()
    setattr(cat, "super_task_category_id", new_pid)
    db.commit()


def get_category_id_by_name_and_parent_id(db: Session, name: str, parent_id: int):
    category = db.query(models.TaskCategory.task_category_id).\
        filter(
            and_(
                models.TaskCategory.name == name,
                models.TaskCategory.super_task_category_id == parent_id
            )
        ).first()
    if not category:
        return None
    return category[0]


def get_category_name_by_id(db: Session, cat_id: int):
    if cat_id is None:
        return "No Category"
    
    cat = db.query(models.TaskCategory.name).\
        filter(models.TaskCategory.task_category_id == cat_id).\
        first()
    return cat[0] if cat else None


def get_subcategories_by_parent_id(db: Session, pid: int):
    subcats = db.query(models.TaskCategory).\
        filter(
            and_(
                models.TaskCategory.super_task_category_id == pid)    
            ).all()
    return subcats


def get_subcategory_ids_by_parent_id(db: Session, pid: int):
    sub_cats = get_subcategories_by_parent_id(db, pid)
    sub_cats_id = [sub.task_category_id for sub in sub_cats]
    return sub_cats_id


def build_category_tree(db: Session, cat_id):
    subs = get_subcategories_by_parent_id(db, cat_id)
    if subs is None: # Base case
        return {}
    res = defaultdict(dict)
    for sub in subs:
        res[sub.name]['subcategories'] = build_category_tree(db, sub.task_category_id)
        res[sub.name]['color'] = sub.color
    return res


def is_super_category(db: Session, super: str, sub: str):
    sub_cat = db.query(models.TaskCategory).\
        filter(
            and_(
                models.TaskCategory.name == sub,
                models.TaskCategory.super_category_name == super
            )
        ).first()
    return sub_cat


def get_last_cat_id(db: Session, cats: list):
    '''
    ex) ['HKU', 'COMP3230', 'Assignment'] -> return id of Assignment
    cats: list[str]
    Return
        parent_id (int): closest parent id
        cumul_path (str): cumulative path from first and second to last category, connected by /
    '''
    parent_id = None
    cumul_path = ""
    if not cats:
        return parent_id, cumul_path

    for cat_name in cats:
        cat_id = get_category_id_by_name_and_parent_id(db, cat_name, parent_id)
        if not cat_id:
            raise exceptions.SubcategoryNotExistException(cumul_path, cat_name)
        parent_id = cat_id
        cumul_path += cat_name + "/"
    return parent_id, cumul_path


def check_exist_category(db: Session, cats: list):
    parent_id = None
    for cat_name in cats:
        cat_id = get_category_id_by_name_and_parent_id(db, cat_name, parent_id)
        if not cat_id:
            return False
        parent_id = cat_id
    return True
    

def get_category_color_by_id(db: Session, cat_id: int):
    cat = db.query(models.TaskCategory).\
        filter(
                and_(
                    models.TaskCategory.task_category_id == cat_id   
                )
            ).first()
    if not cat:
        return None
    return cat.color


def get_category_full_path_by_id(db: Session, category_id: int):
    if category_id is None:
        return ""
    cat_path = ""
    cat = db.query(models.TaskCategory).\
        filter(
            and_(
                models.TaskCategory.task_category_id == category_id
            )
        ).first()
    cat_name = cat.name
    super_cat_id = cat.super_task_category_id
    if super_cat_id is None: # base case
        return cat_name + "/"
    super_cat_full_path = get_category_full_path_by_id(db, super_cat_id)
    cat_path = super_cat_full_path + cat_name + "/"
    return cat_path


def get_all_category_colors(db: Session):
    colors = db.query(models.TaskCategory.name, models.TaskCategory.color).all()
    colors = {c[0]: c[1] for c in colors}
    return colors

    
def get_root_category_id(db: Session, cat_id: int):
    cat = db.query(models.TaskCategory).\
        filter(
                and_(
                    models.TaskCategory.task_category_id == cat_id   
                )
            ).first()
        
    if not cat:
        raise exceptions.CategoryNotExistException(str(cat_id))
    
    pid = cat.super_task_category_id
    if pid is None:
        return cat_id
    
    return get_root_category_id(db,  pid)





from typing import Union, List
import datetime
from collections import defaultdict

from sqlalchemy.orm import Session
from sqlalchemy import and_, func

import girok.server.src.task.models as models
import girok.server.src.task.exceptions as exceptions
import girok.server.src.category.service as category_service
import girok.server.src.utils as general_utils



def create_task(db: Session, task_data):
    new_task = models.Task(**task_data)
    db.add(new_task)
    db.commit()
    db.refresh(new_task)
    return general_utils.sql_obj_to_dict(new_task)


def get_single_task(db: Session, task_id: int):
    task = db.query(models.Task).\
        filter(
            and_(
                models.Task.task_id == task_id
            )
        ).first()
    
    if not task:
        raise exceptions.TaskNotFoundException(task_id=task_id)
    
    task = general_utils.sql_obj_to_dict(task)
    return task


def get_tasks_by_category(
    db: Session,
    cat_ids: List[int],
    start_date: datetime,
    end_date: datetime,
    tag: str,
    priority: Union[int, None] = None,
    no_priority: bool = False,
):
    tasks = defaultdict(dict)
    for cat_id in cat_ids:
        cat_name = category_service.get_category_name_by_id(db, cat_id)
        cat_tasks = get_direct_tasks_of_category(
            db=db,
            cat_id=cat_id,
            start_date=start_date,
            end_date=end_date,
            priority=priority,
            no_priority=no_priority,
            tag=tag
        )
        if cat_id is not None:
            sub_cat_ids = category_service.get_subcategory_ids_by_parent_id(db, cat_id)
        else:
            sub_cat_ids = []
        tasks[cat_name]["tasks"] = cat_tasks
        tasks[cat_name]["sub_categories"] = get_tasks_by_category(
            db=db,
            cat_ids=sub_cat_ids,
            start_date=start_date,
            end_date=end_date,
            priority=priority,
            no_priority=no_priority,
            tag=tag
        )
    
    return tasks
     

def get_tasks_as_list(
    db: Session,
    cat_ids: Union[List[int], None],
    start_date: datetime,
    end_date: datetime,
    tag: str,
    priority: Union[int, None] = None,
    no_priority: bool = False
):
    if not cat_ids:
        return []
    tasks = []
    for cat_id in cat_ids:
        cat_tasks = get_direct_tasks_of_category(
            db=db,
            cat_id=cat_id,
            start_date=start_date,
            end_date=end_date,
            tag=tag,
            priority=priority,
            no_priority=no_priority
        )
        tasks += cat_tasks
        if cat_id is not None:
            sub_cat_ids = category_service.get_subcategory_ids_by_parent_id(db, cat_id)
        else:
            sub_cat_ids = []
        tasks += get_tasks_as_list(
            db=db,
            cat_ids=sub_cat_ids,
            start_date=start_date,
            end_date=end_date,
            tag=tag,
            priority=priority,
            no_priority=no_priority
        )
    
    return tasks
    

def get_tags(db: Session):
    tags = db.query(models.Task.tag).\
        filter(
            and_(
                models.Task.tag != None
            )    
        ).all()
    tags = {tag[0] for tag in tags} # unique tags
    return list(tags)


def get_direct_tasks_of_category(
    db: Session,
    cat_id: int,
    start_date: datetime,
    end_date: datetime,
    tag: str,
    priority: Union[int, None] = None,
    no_priority: bool = False
    ):
    cat_color = None
    if cat_id:
        root_id = category_service.get_root_category_id(db, cat_id)
        cat_color = category_service.get_category_color_by_id(db, root_id)
        
    tasks_query = db.query(models.Task).\
        filter(
            and_(
                models.Task.task_category_id == cat_id,
                func.date(models.Task.deadline) >= start_date,
                func.date(models.Task.deadline) <= end_date
            )
        )
        
    if no_priority:
        tasks_query = tasks_query.filter(models.Task.priority == None)
    else:
        if priority:
            tasks_query = tasks_query.filter(models.Task.priority == priority)
    
    if tag:
        tasks_query = tasks_query.filter(models.Task.tag == tag)
    
    tasks = tasks_query.order_by(models.Task.deadline.asc()).order_by(models.Task.priority.desc()).all()
    tasks_obj_list = general_utils.sql_obj_list_to_dict_list(tasks)
    for task_obj in tasks_obj_list:
        task_obj.update({
            "task_category_full_path": category_service.get_category_full_path_by_id(db, task_obj['task_category_id']),
        })
        
        if cat_color:
            task_obj.update({"color": cat_color})

    return tasks_obj_list
    
    
def delete_task(db: Session, task_id: int):
    task = db.query(models.Task).\
        filter(
            and_(
                models.Task.task_id == task_id
            )
        ).first()
    if not task:
        raise exceptions.TaskNotFoundException(task_id=task_id)
        
    db.delete(task)
    db.commit()
    return task


def change_task_tag(db: Session, task_id: int, new_tag_name: str):
    task = db.query(models.Task).\
        filter(
            and_(
                models.Task.task_id == task_id
            )
        ).first()
        
    if not task:
        raise exceptions.TaskNotFoundException(task_id=task_id)

    setattr(task, "tag", new_tag_name)
    db.commit()
    return general_utils.sql_obj_to_dict(task)
    


def change_task_priority(db: Session, task_id: int, new_priority: int):
    task = db.query(models.Task).\
        filter(
            and_(
                models.Task.task_id == task_id
            )
        ).first()
        
    if not task:
        raise exceptions.TaskNotFoundException(task_id=task_id)

    setattr(task, "priority", new_priority)
    db.commit()
    return general_utils.sql_obj_to_dict(task)

    
def change_task_date(db: Session, task_id: int, new_date: str):
    task = db.query(models.Task).\
        filter(
            and_(
                models.Task.task_id == task_id
            )
        ).first()
        
    if not task:
        raise exceptions.TaskNotFoundException(task_id=task_id)

    setattr(task, "deadline", new_date)
    db.commit()
    return general_utils.sql_obj_to_dict(task)


def change_task_name(db: Session, task_id: int, new_name: str):
    task = db.query(models.Task).\
        filter(
            and_(
                models.Task.task_id == task_id
            )
        ).first()
        
    if not task:
        raise exceptions.TaskNotFoundException(task_id=task_id)

    setattr(task, "name", new_name)
    db.commit()
    return general_utils.sql_obj_to_dict(task)



import datetime
from collections import Counter
from typing import List


def get_current_year():
    now = datetime.datetime.now()
    return now.year


def extract_sub_dict(dct: dict, included_keys: List[str]):
    return {k: v for k, v in dct.items() if k in included_keys}


def capitalize_variable(var_name: str):
    """
    capitalizing a variable

    e.g. a_bcD -> ABcD

    :param var_name: name of variable in string
    :return: capitalized name
    """

    return ''.join([str(_).capitalize() for _ in var_name.split('_')])


def count_enum(values, enumTy, asec=False):
    """
    Count enum in list
    :param values: list of enum value
    :param enumTy: enum type
    :return: { enum key: count }
    """
        
    return {k.value: 0 for k in enumTy} | Counter(values)


def dict_as_list(dct: dict, asc=False):
    keys = list(dct.keys())
    values = list(dct.values())
    if not asc:
        keys = keys[::-1]
        values = values[::-1]
    return {"keys": keys, "values": values}


def sql_obj_list_to_dict_list(sql_obj_list):
    return [sql_obj_to_dict(sql_obj) for sql_obj in sql_obj_list]


def sql_obj_to_dict(sql_obj):
    return {col.name: getattr(sql_obj, col.name) for col in sql_obj.__table__.columns}



        
    
        


from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
from girok.config import get_config

cfg = get_config()

SQLITE_DATABASE_URL = f"sqlite:///{cfg.app_dir}/database.db"
engine = create_engine(SQLITE_DATABASE_URL)
SessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=engine)
Base = declarative_base()

def get_db():
    db = SessionLocal()
    try:
        yield db
    finally:
        db.close()


import requests
from rich.console import Console

from girok.config import get_config
import girok.utils.auth as auth_utils
import girok.utils.general as general_utils
import girok.utils.display as display_utils
import girok.constants as constants

# Guest mode imports
import girok.server.src.category.router as category_router

console = Console()
cfg = get_config()

def get_categories():
    mode = auth_utils.get_mode(cfg.config_path)
    if mode == "user":
        resp = requests.get(
            cfg.base_url + "/categories",
            headers=auth_utils.build_jwt_header(cfg.config_path)
        )
        if resp.status_code == 200:
            return general_utils.bytes2dict(resp.content)
    elif mode == "guest":
        resp = category_router.get_all_categories()
        return resp
    

def add_category(cat_str: str, color=None):
    mode = auth_utils.get_mode(cfg.config_path)
    cats = cat_str.split('/')
    if mode == "user":
        resp = requests.post(
            cfg.base_url + "/categories",
            json={
                "names": cats,
                "color": color
            },
            headers=auth_utils.build_jwt_header(cfg.config_path)
        )
        if resp.status_code == 201:
            display_utils.center_print("Category added successfully!", type="success")
            cats_dict = get_categories()
            display_utils.display_categories(cats_dict, highlight_cat=cat_str)
        elif resp.status_code == 400:
            err_msg = general_utils.bytes2dict(resp.content)['detail']
            display_utils.center_print(err_msg, type="error")
        else:
            print(resp)
        return resp
    elif mode == "guest":
        resp = category_router.create_category({"names": cats, "color": color})
        if resp['success']:
            display_utils.center_print("Category added successfully!", type="success")
            cats_dict = get_categories()
            display_utils.display_categories(cats_dict, highlight_cat=cat_str)
        else:
            display_utils.center_print(resp['detail'], type="error")
        
        
def remove_category(cat_str: str):
    mode = auth_utils.get_mode(cfg.config_path)
    cats = cat_str.split('/')
    if mode == "user":
        resp = requests.delete(
            cfg.base_url + "/categories",
            json={
                "cats": cats
            },
            headers=auth_utils.build_jwt_header(cfg.config_path)
        )
        if resp.status_code == 204:
            display_utils.center_print(f"Deleted {cat_str} successfully.", type="success")
            cats_dict = get_categories()
            display_utils.display_categories(cats_dict)
        elif resp.status_code == 400:
            err_msg = general_utils.bytes2dict(resp.content)['detail']
            display_utils.center_print(err_msg, type="error")
        else:
            display_utils.center_print(resp.content, type="error")
    elif mode == "guest":
        resp = category_router.delete_category({"cats": cats})
        if resp['success']:
            display_utils.center_print(f"Deleted {cat_str} successfully.", type="success")
            cats_dict = get_categories()
            display_utils.display_categories(cats_dict)
        else:
            display_utils.center_print(resp['detail'], type="error")
            
        
def rename_category(cat_str: str, new_name: str):
    mode = auth_utils.get_mode(cfg.config_path)
    cats = cat_str.split('/')
    
    if mode == "user":
        resp = requests.patch(
            cfg.base_url + "/categories/name",
            json={
                "cats": cats,
                "new_name": new_name
            },
            headers=auth_utils.build_jwt_header(cfg.config_path)
        )
        if resp.status_code == 204:
            new_cat = '/'.join(cat_str.split('/')[:-1] + [new_name])
            display_utils.center_print(f"Successfully renamed {cat_str} to {new_cat}.", type="success")
            cats_dict = get_categories()
            display_utils.display_categories(cats_dict, highlight_cat=new_cat)
        elif resp.status_code == 400:
            err_msg = general_utils.bytes2dict(resp.content)['detail']
            display_utils.center_print(err_msg, type="error")
        else:
            display_utils.center_print(resp.content, type="error")
    elif mode == "guest":
        resp = category_router.rename_category({
            "cats": cats,
            "new_name": new_name
        })
        if resp['success']:
            new_cat = '/'.join(cat_str.split('/')[:-1] + [new_name])
            display_utils.center_print(f"Successfully renamed {cat_str} to {new_cat}.", type="success")
            cats_dict = get_categories()
            display_utils.display_categories(cats_dict, highlight_cat=new_cat)
        else:
            display_utils.center_print(resp['detail'], type="error")


def move_category(cat_str: str, new_parent_cat_str: str):
    mode = auth_utils.get_mode(cfg.config_path)
    if cat_str.endswith('/'):
        cat_str = cat_str[:-1]
    if new_parent_cat_str.endswith('/'):
        new_parent_cat_str = new_parent_cat_str[:-1]
        
    cats = cat_str.split('/') if cat_str else []
    new_parent_cats = new_parent_cat_str.split('/') if new_parent_cat_str else []

    if mode == "user":
        resp = requests.patch(
            cfg.base_url + "/categories/parent",
            json={
                "cats": cats,
                "new_parent_cats": new_parent_cats
            },
            headers=auth_utils.build_jwt_header(cfg.config_path)
        )
        if resp.status_code == 200:
            new_cat = '/'.join(new_parent_cat_str.split('/') + [cat_str.split('/')[-1]])
            display_utils.center_print(f"Successfully moved {cat_str} to {new_parent_cat_str}/.", type="success")
            cats_dict = get_categories()
            display_utils.display_categories(cats_dict, highlight_cat=new_cat)
        elif resp.status_code == 400:
            err_msg = general_utils.bytes2dict(resp.content)['detail']
            display_utils.center_print(err_msg, type="error")
        else:
            display_utils.center_print(resp.content, type="error")
        return resp
    elif mode == "guest":
        resp = category_router.move_category({
            "cats": cats,
            "new_parent_cats": new_parent_cats
        })
        if resp['success']:
            new_cat = '/'.join(new_parent_cat_str.split('/') + [cat_str.split('/')[-1]])
            display_utils.center_print(f"Successfully moved {cat_str} to {new_parent_cat_str}/.", type="success")
            cats_dict = get_categories()
            display_utils.display_categories(cats_dict, highlight_cat=new_cat)
        else:
            display_utils.center_print(resp['detail'], type="error") 


def get_last_cat_id(cats: list):
    mode = auth_utils.get_mode(cfg.config_path)
    if mode == "user":
        resp = requests.get(
            cfg.base_url + "/categories/last-cat-id",
            json={
                "cats": cats
            },
            headers=auth_utils.build_jwt_header(cfg.config_path)
        )
        if resp.status_code == 200:
            return general_utils.bytes2dict(resp.content)['cat_id']
        elif resp.status_code == 400:
            err_msg = general_utils.bytes2dict(resp.content)['detail']
            display_utils.center_print(err_msg, type="error")
            exit(0)
        else:
            display_utils.center_print(resp.content, type="error")
            exit(0)
    elif mode == "guest":
        resp = category_router.get_last_cat_id({"cats": cats})
        if resp['success']:
            return resp['cat_id']
        else:
            display_utils.center_print(resp['detail'], type="error")
    exit(0)
            
        
        
def get_category_color(cat_id: int):
    mode = auth_utils.get_mode(cfg.config_path)
    if mode == "user":
        resp = requests.get(
            cfg.base_url + f"/categories/{cat_id}/color",
            headers=auth_utils.build_jwt_header(cfg.config_path)
        )
        
        if resp.status_code == 200:
            return general_utils.bytes2dict(resp.content)['color']
        elif resp.status_code == 400:
            err_msg = general_utils.bytes2dict(resp.content)['detail']
            display_utils.center_print(err_msg, type="error")
            exit(0)
        else:
            display_utils.center_print(resp.content, type="error")
            exit(0)
    elif mode == "guest":
        resp = category_router.get_category_color(cat_id)
        if resp['success']:
            return resp['color']
        else:
            display_utils.center_print(resp.content, type="error")
    exit(0)
    
def get_color_dict():
    mode = auth_utils.get_mode(cfg.config_path)
    if mode == "user":
        resp = requests.get(
            cfg.base_url + "/categories/color",
            headers=auth_utils.build_jwt_header(cfg.config_path)
        )
        if resp.status_code == 200:
            color_dict = general_utils.bytes2dict(resp.content)
            return color_dict
        elif resp.status_code == 400:
            err_msg = general_utils.bytes2dict(resp.content)['detail']
            display_utils.center_print(err_msg, type="error")
        else:
            display_utils.center_print("Error occurred.", type="error")
    elif mode == "guest":
        resp = category_router.get_category_colors_dict()
        if resp['success']:
            return resp['colors']    
        else:
            display_utils.center_print("Error occurred.", type="error")
    exit(0)
            


from sqlalchemy import Column, String, Integer, ForeignKey
from sqlalchemy.sql.sqltypes import TIMESTAMP, Boolean, DateTime
from sqlalchemy.sql.expression import text
from sqlalchemy.orm import relationship

from girok.server.src.database import Base
    

class TaskCategory(Base):
    __tablename__ = "task_category"
    task_category_id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String(50), nullable=False)
    super_task_category_id = Column(Integer, ForeignKey('task_category.task_category_id'), nullable=True)
    color = Column(String(20), nullable=True)
    children = relationship('TaskCategory', cascade="all,delete")




from sqlalchemy import Column, String, Integer, ForeignKey
from sqlalchemy.sql.sqltypes import TIMESTAMP, Boolean, DateTime
from typing import Union, List, Dict
from pydantic import BaseModel, Field, validator
from datetime import datetime, timedelta
import girok.server.src.task.enums as task_enums

class TaskCreateIn(BaseModel):
    task_category_id: Union[int, None] = None
    name: str
    deadline: str
    priority: int = Field(default=None, ge=1, le=5)
    color: Union[str, None] = None
    everyday: Union[bool, None] = False
    tag: Union[str, None] = None
    is_time: bool = False
    all_day: Union[bool, None] = False
    weekly_repeat: int = Field(default=None, gt=0, le=6)


class TaskCreateOut(BaseModel):
    task_id: int
    
    class Config:
        orm_mode = True
        

class Task(BaseModel):
    task_id: int
    user_id: int
    task_category_id: Union[int, None]
    category_path: Union[str, None]
    name: str
    color: str
    deadline: str
    all_day: Union[bool, None]
    is_time: bool
    priority: Union[int, None]
    everyday: Union[bool, None]
    created_at: datetime
    
    class Config:
        orm_mode = True
        

class TaskGetIn(BaseModel):
    category: Union[List[str], None] = Field(
        default=None,
        title="Category",
        description="Full category path of a category. Ex. ['HKU', 'COMP3230'].",
    ),
    start_date: Union[str, None] = Field(
        default="2000-01-01 00:00:00",
        title="Start date",
        description="Start date. Ex. '2023-01-23 12:00:00'",
        regex="^[0-9]{4}-[0-9]{1,2}-[0-9]{1,2} [0-9]{2}:[0-9]{2}:[0-9]{2}$"
    ),
    end_date: Union[str, None] = Field(
        default=(datetime.now() + timedelta(days=365*10)).strftime("%Y-%m-%d 00:00:00"),
        title="End date",
        description="End date. Ex. '2023-03-23 12:00:00'",
        regex="^[0-9]{4}-[0-9]{1,2}-[0-9]{1,2} [0-9]{2}:[0-9]{2}:[0-9]{2}$"
    ),
    priority: Union[int, None] = Field(
        default=None,
        title="Priority",
        ge=1,
        le=5
    ),
    no_priority: bool = False,
    tag: Union[str, None] = Field(
        default_factory=None,
        title="Tag"
    ),
    view: Union[task_enums.TaskView, None]= Field(
        default_factory=None,
        title="View method for tasks"  
    )
    
    @validator("start_date")
    def set_start_date(cls, v):
        if v is None:
            return "2000-01-01 00:00:00"
        return v

    @validator("end_date")
    def set_end_date(cls, v):
        if v is None:
            return (datetime.now() + timedelta(days=365*10)).strftime("%Y-%m-%d 00:00:00")
        return v
    

class GetSingleTaskOut(BaseModel):
    name: str
    
        
class TaskOut(BaseModel):
    tasks: Dict[str, dict]
    
    
class Tag(BaseModel):
    tag_id: int
    task_id: int
    name: int
    
class TagOut(BaseModel):
    tags: List[str]
    

class ChangeTaskTagIn(BaseModel):
    new_tag_name: str


class ChangeTaskPriorityIn(BaseModel):
    new_priority: int = Field(ge=1, le=5)


class ChangeTaskDateIn(BaseModel):
    new_date: str = Field(default=..., regex="^([0-9]){4}-([0-9]){1,2}-([0-9]){1,2} [0-9]{2}:[0-9]{2}:[0-9]{2}$")


class ChangeTaskNameIn(BaseModel):
    new_name: str


from pathlib import Path
import os.path as osp
import os
import time

import typer
from rich import print
from rich.console import Console
from rich.table import Table
from rich.progress import Progress

from girok.config import get_config
import girok.utils.general as general_utils
import girok.utils.auth as auth_utils
import girok.api.auth as auth_api
import girok.utils.display as display_utils

import girok.server.src.task.models as task_models
import girok.server.src.category.models as category_models
from girok.server.src.database import engine

app = typer.Typer(rich_markup_mode='rich')

cfg = get_config()
console=Console()


@app.command("mode", help="Current mode", rich_help_panel=":lock: [bold yellow1]Authentication Commands[/bold yellow1]")
def mode():
    mode = auth_utils.get_mode(cfg.config_path)
    if mode == "off":
        print("You're not logged in.")
    elif mode == "guest":
        print("[yellow]Guest mode[/yellow]")
    elif mode == "user":
        print("[yellow]User mode[/yellow]")
        

@app.command("guest", help="[yellow]Login as a guest user (using local storage)[/yellow]", rich_help_panel=":lock: [bold yellow1]Authentication Commands[/bold yellow1]")
def guest():
    if not os.path.exists(cfg.app_dir + "/database.db"):
        task_models.Base.metadata.create_all(bind=engine)
        category_models.Base.metadata.create_all(bind=engine)
    general_utils.update_json(cfg.config_path, {"mode": "guest"})
    print("[yellow]You're logged in as a guest![/yellow]")
    

@app.command("login", help="[yellow]Login[/yellow] with email and password", rich_help_panel=":lock: [bold yellow1]Authentication Commands[/bold yellow1]")
def login():
    # Check if the user holds a valid JWT (logged in)
    stored_access_token = auth_utils.get_access_token_from_json(cfg.config_path)
    if stored_access_token:
        resp = auth_api.validate_access_token(stored_access_token)
        if resp.status_code == 200:
            print("You have already logged in. Enjoy!")
            exit(0)
    
    # Log-in
    print("Please login to proceed\n")
    email = typer.prompt("Enter email address")
    password = typer.prompt("Enter password", hide_input=True)
    
    resp = auth_api.login(email, password)
    if resp.status_code == 200:
        access_token = general_utils.bytes2dict(resp.content)['access_token']
        general_utils.update_json(cfg.config_path, {"access_token": access_token, "mode": "user"})
        print("[yellow]You're logged in as a user![/yellow]")
    elif resp.status_code == 401:
        err_msg = general_utils.bytes2dict(resp.content)['detail']
        display_utils.center_print(err_msg, type="error")
        exit(0)
    else:
        print("Login failed. Please try again with [green]girok login[/green]")
        exit(0)
    exit(0)
    

@app.command("logout", help="[red]Logout[/red] from the currently logged-in account", rich_help_panel=":lock: [bold yellow1]Authentication Commands[/bold yellow1]")
def logout():
    mode = auth_utils.get_mode(cfg.config_path)
    if mode == "guest":
        auth_utils.log_out_from_guest(cfg.config_path)
        print("Successfully logged out from [yellow]guest mode[/yellow].")
    elif mode == "user":
        auth_utils.log_out_from_user(cfg.config_path)
        # auth_utils.remove_access_token(cfg.config_path)
        print("Successfully logged out from [yellow]user mode[/yellow].")
        exit(0)
    
    
@app.command("register", help="[green]Register[/green] a new account", rich_help_panel=":lock: [bold yellow1]Authentication Commands[/bold yellow1]")
def register():
    access_token = auth_utils.get_access_token_from_json(cfg.config_path)
    if access_token and auth_utils.is_logged_in(access_token):
        print("You're logged in. Please try after log out.")
        exit(0)
        
    print(":star: Welcome to [yellow]girok[/yellow]! :star:\n")
    is_register = typer.confirm("Do you want to register a new account?")
    if is_register:
        email = typer.prompt("\nEnter email address")
        password = typer.prompt("Enter password", hide_input=True)
        confirm_password = typer.prompt("Confirm password", hide_input=True)
        auth_utils.match_passwords(password, confirm_password)
        
        # register a new account
        register_resp = auth_api.register(email, password)
        if register_resp.status_code == 201:
            print("\nPlease enter the [yellow]verification code[/yellow] sent to your email.")
            print("If you don't find it in your inbox, please check [red]spam/junk[/red] email.\n")
            cnt_try = 3
            success = False
            while cnt_try > 0:
                verification_code = typer.prompt(f"Verification code [{cnt_try} tries left]")
                if auth_api.verify_verification_code(email, verification_code): 
                    success = True
                    break
                else:
                    cnt_try -= 1
            if not success:
                display_utils.center_print("Verification failed. Please register again.", type="error")
                exit(0)
        elif register_resp.status_code == 400:
            err_msg = general_utils.bytes2dict(register_resp.content)['detail']
            display_utils.center_print(err_msg, type="error")
            exit(0)
        else:
            err_msg = general_utils.bytes2dict(register_resp.content)['detail'][0]['msg']
            display_utils.center_print(str(err_msg), type="error")
            exit(0)
            
        display_utils.register_welcome()
        exit(0)
    else:
        exit(0)

if __name__ == '__main__':
    app()


from pathlib import Path
import os
import os.path as osp

import typer
from rich import print
from rich.console import Console
from rich.table import Table

from girok.config import get_config
import girok.api.auth as auth_api
import girok.commands.auth as auth_command
import girok.commands.category as category_command
import girok.commands.task as task_command
import girok.commands.calendar as calendar_command
import girok.commands.info as info_command
import girok.utils.general as general_utils
import girok.utils.auth as auth_utils
import girok.utils.display as display_utils

app = typer.Typer(rich_markup_mode='rich', help="Enter [red]girok <command name> --help[/red] to see more detailed documentations of commands!")
app.registered_commands.extend(auth_command.app.registered_commands)
app.registered_commands.extend(category_command.app.registered_commands)
app.registered_commands.extend(task_command.app.registered_commands)
app.registered_commands.extend(calendar_command.app.registered_commands)

cfg = get_config()

def version_callback(value: bool):
    if value:
        print(f"[yellow]{cfg.version}[/yellow]")
        raise typer.Exit()
    
    
####### Major refactoring needed -> minimize code duplications for local-storage feature
@app.callback()
def pre_command_callback(ctx: typer.Context, version: bool = typer.Option(None, "--version", callback=version_callback, is_eager=True),):
    cmd = ctx.invoked_subcommand
    
    # Setting up app dir and config file if they don't exist
    general_utils.config_setup(cfg.app_dir, cfg.config_path)

    # Authentication status check
    if cmd == "guest":
        # Check if logged-out
        mode = auth_utils.get_mode(cfg.config_path)
        if mode != "off":
            print(f"You're already logged in as [yellow]{mode}[/yellow]. Please log out and try again.")
            exit(0)
        return
    elif cmd == "login":
        # Check if logged-out
        mode = auth_utils.get_mode(cfg.config_path)
        if mode != "off":
            print(f"You're already logged in as [yellow]{mode}[/yellow]. Please log out and try again.")
            exit(0)
        return
    elif cmd == "logout":
        # Check if logged-in
        mode = auth_utils.get_mode(cfg.config_path)
        if mode not in ["guest", "user"]:
            print("You're not logged in.")
            exit(0)
        return
    elif cmd == "register":
        mode = auth_utils.get_mode(cfg.config_path)
        if mode != "off":
            print(f"You're already logged in as [yellow]{mode}[/yellow]. Please log out and try again.")
            exit(0)
        return
    elif cmd == "mode":
        return
    else: # non-auth commands
        # Check if logged-in
        mode = auth_utils.get_mode(cfg.config_path)
        if mode == "off":
            print("You're not logged in. Please login with [yellow]girok login[/yellow] or [yellow]girok guest[/yellow].")
            exit(0)
        elif mode == "guest":
            pass
        elif mode == "user":
            if not auth_utils.is_logged_in_as_user(cfg.config_path):
                print("You're not logged in. Please login with [yellow]girok login[/yellow] or [yellow]girok guest[/yellow].")
                exit(0)
        else:
            raise Exception("Unexpected Authentication")
        


from typing import List
from collections import defaultdict
from sqlalchemy.orm import Session

from girok.server.src.database import get_db
import girok.server.src.category.schemas as schemas
import girok.server.src.category.service as service
import girok.server.src.category.exceptions as exceptions
import girok.server.src.category.constants as category_constants


def get_all_categories():
    db = next(get_db())
    resp = defaultdict(dict)
    cats = service.get_subcategories_by_parent_id(db, pid=None)
    for cat in cats:
        resp[cat.name]["subcategories"] = service.build_category_tree(db, cat.task_category_id)
        resp[cat.name]["color"] = cat.color
        
    resp['No Category'] = {
        "color": "grey",
        "subcategories": {}
    }
    
    return resp


def create_category(category: schemas.CategoryCreateIn):
    db = next(get_db())
    ''' Example
    category = {
        names: ['HKU', 'COMP3230', 'Assignment']
        color: yellow
    }
    '''
    names = category['names']
    color = category['color']
    if color and color not in category_constants.CATEGORY_COLORS:
        return {"success": False, "detail": exceptions.CategoryColorNotExistException(color).detail}
    
    try:
        pid_of_second_last_cat, cumul_path = service.get_last_cat_id(db, names[:-1])
    except Exception as e:
        return {"success": False, "detail": e.detail}
    
    # If given, check if the subdirectory color is equal to the parent's color,
    if color is None:
        if pid_of_second_last_cat is not None:
            color = service.get_category_color_by_id(db, pid_of_second_last_cat) # set to parent's color
        else:
            # Select non-overlapping colors
            existing_colors = set(service.get_all_category_colors(db).values())
            for c in category_constants.CATEGORY_COLORS: # Assign color (lower index higher priority)
                if c in existing_colors: 
                    continue
                color = c
                break
            if color is None: # If no available non-overlapping color, assign the default color
                color = category_constants.DEFAULT_CATEGORY_COLOR
    else:
        if pid_of_second_last_cat is not None:
            parent_color = service.get_category_color_by_id(db, pid_of_second_last_cat) 
            if color != parent_color:
                return {"success": False, "detail": exceptions.CategoryColorException(names[-1], color, '/'.join(names[:-1]), parent_color).detail}
            
    
    new_cat_name = names[-1]
    dup_cat_id = service.get_category_id_by_name_and_parent_id(db, new_cat_name, pid_of_second_last_cat)
    if dup_cat_id:
        return {"success": False, "detail": exceptions.CategoryAlreadyExistsException(cumul_path, new_cat_name).detail}
    
    new_cat_data = {
        "name": new_cat_name,
        "super_task_category_id": pid_of_second_last_cat,
        "color": color
    }
    new_cat = service.create_category(db, new_cat_data)
    return {"success": True, "data": new_cat}


def delete_category(category: schemas.CategoryDeleteIn):
    db = next(get_db())
    cats = category['cats']
    try:
        cat_id, _ = service.get_last_cat_id(db, cats)
    except Exception as e:
        return {"success": False, "detail": e.detail}
    
    service.delete_category(db, cat_id)
    return {"success": True}
    
        
def rename_category(category: schemas.CategoryRenameIn):
    db = next(get_db())
    cats, new_name, = category['cats'], category['new_name']
    
    # Get cat_id of the original category
    try:
        cat_id, _ = service.get_last_cat_id(db, cats)
    except Exception as e:
        return {"success": False, "detail": e.detail}
    
    # Check if new_cats exists (ex. cats[:-1] + /new_cat)
    new_cats = cats[:-1] + [new_name]
    if service.check_exist_category(db, new_cats):
        return {"success": False, "detail": exceptions.CategoryAlreadyExistsException('/'.join(new_cats[:-1]), new_name)}
    service.rename_category(db, cat_id, new_name)
    return {"success": True}
    

def move_category(category: schemas.CategoryMoveIn):
    """
    cats: ['HKU', 'COMP3230']
    new_parent_cats: ['Dev', 'Git']
    
    Then, we want to move the entire HKU/COMP3230 category into under Dev/Git category.
    """
    db = next(get_db())
    cats, new_parent_cats = category['cats'], category['new_parent_cats']
    if not cats:
        return {"success": False, "detail": exceptions.CannotMoveRootDirectoryException().detail}
    
    new_cat = new_parent_cats + [cats[-1]]
    
    if service.check_exist_category(db, new_parent_cats + [cats[-1]]):
        raise exceptions.CategoryAlreadyExistsException('/'.join(new_parent_cats), cats[-1])
        
    try:
        cat_id, _ = service.get_last_cat_id(db, cats)
        new_pid, _ = service.get_last_cat_id(db, new_parent_cats)
    except Exception as e:
        return {"success": False, "detail": e.detail}
    if cat_id == new_pid:
        return {"success": False, "detail": exceptions.CannotMoveToSameLocation().detail}
    
    service.move_category(db, cat_id, new_pid)
    return {"success": True}
    
    
def get_last_cat_id(cat_data: schemas.LastCategoryIdIn):
    db = next(get_db())
    try:
        cat_id, _ = service.get_last_cat_id(db, cat_data['cats'])
        return {"success": True, "cat_id": cat_id}
    except Exception as e:
        return {"success": False, "detail": e.detail}


def get_category_color(cat_id: int):
    db = next(get_db())
    try:
        color = service.get_category_color_by_id(db, cat_id)
        return {"success": True, "color": color}
    except Exception as e:
        return {"success": False, "detail": e.detail}


def get_category_colors_dict():
    db = next(get_db())
    try:
        colors = service.get_all_category_colors(db)
        colors['No Category'] = "grey"
        return {"success": True, "colors": colors}
    except Exception as e:
        return {"success": False, "detail": e.detail}


import girok.server.src.category.constants as category_constants


class CategoryAlreadyExistsException(Exception):
    def __init__(self, sup, sub):
        if sup == "":
            self.detail = f"There already exists a category named '{sub}'."
        else:
            self.detail = f"'{sup}' already has a subcategory named '{sub}'"
        

class CategoryNotExistException(Exception):
    def __init__(self, name):
        self.detail = f"Category '{name}' does not exist."
        
        
class SubcategoryNotExistException(Exception):
    def __init__(self, sup, sub):
        if sup == "":
            self.detail = f"There is no category named {sub}."
        else:
            self.detail = f"'{sup}' does not have a subcategory named '{sub}'"
            

class CannotMoveRootDirectoryException(Exception):
    def __init__(self):
        self.detail = "Cannot move the root directory!"
        
        
class CannotMoveToSameLocation(Exception):
    def __init__(self):
        self.detail = "[Warning] Circular reference prohibited."
        
        
class CategoryColorException(Exception):
    def __init__(self, cat_path, color, parent_cat_path, parent_color):
        self.detail = f"You have given {color} color for {cat_path}. However, {parent_cat_path} already has a color {parent_color}"
        
class CategoryColorNotExistException(Exception):
    def __init__(self, color):
        self.detail = f"'{color}' is not a valid color. Please choose from {category_constants.CATEGORY_COLORS}."



class InvalidPriorityWindowException(Exception):
    def __init__(self):
        self.detail = "Minimum priority must be less than or equal to maximum priority."
        
    
class InvalidDateWindowException(Exception):
    def __init__(self, start, end):
        self.detail = f"Start date {start} must be before or equal to end date {end}."
        
        
class InvalidPriorityPairException(Exception):
    def __init__(self):
        self.detail = f"Min and max priority must be both present or all None."
        
class TaskNotFoundException(Exception):
    def __init__(self, task_id):
        self.detail = f"Task id {task_id} is not found."




import typer

app = typer.Typer(rich_markup_mode='rich')

@app.command("version", help="Show version", rich_help_panel=":tear-off_calendar: [bold yellow1]Info Commands[/bold yellow1]")
def show_calendar():
    print("hello")




from typing import Union, List, Dict
from pydantic import BaseModel
    

class CategoryIn(BaseModel):
    user_email: str

class CategoryOut(BaseModel):
    resp: Dict[str, dict]
    
class CategoryCreateIn(BaseModel):
    color: Union[str, None]
    names: List[str]
    

class CategoryCreateOut(BaseModel):
    task_category_id: int
    class Config:
        orm_mode = True
        
        
class CategoryDeleteIn(BaseModel):
    cats: List[str]
    

class CategoryRenameIn(BaseModel):
    cats: List[str]
    new_name: str
    
class CategoryMoveIn(BaseModel):
    cats: List[str]
    new_parent_cats: List[str]
    
class LastCategoryIdIn(BaseModel):
    cats: List[str]
    
    


from typing import Union
from textual import log
from datetime import datetime
import requests

from girok.config import get_config
import girok.utils.auth as auth_utils
import girok.utils.display as display_utils
import girok.utils.general as general_utils
import girok.utils.task as task_utils
import girok.constants as constants

import girok.server.src.task.router as task_router

cfg = get_config()

def create_task(task_data: dict):
    mode = auth_utils.get_mode(cfg.config_path)
    if mode == "user":
        resp = requests.post(
            cfg.base_url + "/tasks",
            json=task_data,
            headers=auth_utils.build_jwt_header(cfg.config_path)
        )
        if resp.status_code == 201:
            task = general_utils.bytes2dict(resp.content)
            task_id = task['task_id']
            return task_id
        elif resp.status_code == 400:
            err_msg = general_utils.bytes2dict(resp.content)['detail']
            display_utils.center_print(err_msg, constants.DISPLAY_TERMINAL_COLOR_ERROR)
            exit(0)
        else:
            display_utils.center_print("Error occurred.", constants.DISPLAY_TERMINAL_COLOR_ERROR)
            exit(0)
    elif mode == "guest":
        resp = task_router.create_task(task_data)
        if resp['success']:
            task = resp['new_task']
            task_id = task['task_id']
            return task_id
        else:
            display_utils.center_print(resp['detail'], type="error")
            exit(0)
    exit(0) 


def get_single_task(task_id: int):
    mode = auth_utils.get_mode(cfg.config_path)
    if mode == "user":
        resp = requests.get(
            cfg.base_url + f"/tasks/{task_id}",
            headers=auth_utils.build_jwt_header(cfg.config_path),
        )
        if resp.status_code == 200:
            task = general_utils.bytes2dict(resp.content)
            return task
        elif resp.status_code == 400:
            err_msg = general_utils.bytes2dict(resp.content)['detail']
            display_utils.center_print(err_msg, type="error")
        else:
            display_utils.center_print(resp.content, type="error")
    elif mode == "guest":
        resp = task_router.get_single_task(task_id)
        if resp['success']:
            task = resp['task']
            return task
        else:
            display_utils.center_print(resp['detail'], type="error")
    exit(0)


def get_tasks(
    cats: Union[list, None] = None,
    start_date: Union[str, None] = None,
    end_date: Union[str, None] = None,
    tag: Union[str, None] = None,
    priority: Union[int, None] = None,
    no_priority: bool = False,
    view: str = "category"
):
    mode = auth_utils.get_mode(cfg.config_path)
    query_str_obj = {
        "category": cats,
        "start_date": start_date,
        "end_date": end_date,
        "tag": tag,
        "priority": priority,
        "no_priority": no_priority,
        "view": view
    }
    if mode == "user":
        resp = requests.get(
            cfg.base_url + "/tasks",
            headers=auth_utils.build_jwt_header(cfg.config_path),
            params=query_str_obj
        )
        if resp.status_code == 200:
            return general_utils.bytes2dict(resp.content)['tasks']
        if resp.status_code == 400:
            err_msg = general_utils.bytes2dict(resp.content)['detail']
            display_utils.center_print(err_msg, type="error")
            exit(0)
        elif resp.status_code:
            display_utils.center_print("Error occurred.", type="title")
            exit(0)
    elif mode == "guest":
        resp = task_router.get_tasks(query_str_obj)
        if resp['success']:
            return resp['tasks']
        else:
            display_utils.center_print(resp['detail'], type="error")
            
    exit(0)


def remove_task(task_id: int):
    mode = auth_utils.get_mode(cfg.config_path)
    if mode == "user":
        resp = requests.delete(
            cfg.base_url + f"/tasks/{task_id}",
            headers=auth_utils.build_jwt_header(cfg.config_path),
        )
        if resp.status_code == 204:
            return True
        elif resp.status_code == 400:
            err_msg = general_utils.bytes2dict(resp.content)['detail']
            display_utils.center_print(err_msg, type="error")
        else:
            display_utils.center_print(resp.content, type="error")
    elif mode == "guest":
        resp = task_router.delete_task(task_id)
        if resp['success']:
            return True
        else:
            display_utils.center_print(resp['detail'], type="error") 
    exit(0)

def get_tags():
    mode = auth_utils.get_mode(cfg.config_path)
    if mode == "user":
        resp = requests.get(
            cfg.base_url + "/tasks/tags",
            headers=auth_utils.build_jwt_header(cfg.config_path)
        )
        if resp.status_code == 200:
            return general_utils.bytes2dict(resp.content)['tags']
        elif resp.status_code == 400:
            err_msg = general_utils.bytes2dict(resp.content)['detail']
            display_utils.center_print(err_msg, type="error")
            exit(0)
        else:
            exit(0)
    elif mode == "guest":
        resp = task_router.get_tags()
        if resp['success']:
            return resp['tags']
        else:
            display_utils.center_print(resp['detail'], type="error")
    exit(0)
            
    
def change_task_tag(task_id: int, new_tag_name: str):
    mode = auth_utils.get_mode(cfg.config_path)
    if mode == "user":
        resp = requests.patch(
            cfg.base_url + f"/tasks/{task_id}/tag",
            headers=auth_utils.build_jwt_header(cfg.config_path),
            json={
                "new_tag_name": new_tag_name
            }
        )
        
        if resp.status_code == 200:
            task = general_utils.bytes2dict(resp.content)
            task_name = task['name']
            display_utils.center_print(f"Successfully changed [{task_name}]'s tag to {new_tag_name}.", type="success")
            return
        elif resp.status_code == 400:
            err_msg = general_utils.bytes2dict(resp.content)['detail']
            display_utils.center_print(err_msg, type="error")
        else:
            display_utils.center_print(resp.content, type="error")
    elif mode == "guest":
        resp = task_router.change_task_tag(task_id, {"new_tag_name": new_tag_name})
        if resp['success']:
            task = resp['updated_task']
            task_name = task['name']
            display_utils.center_print(f"Successfully changed [{task_name}]'s tag to {new_tag_name}.", type="success")
            return
        else:
            display_utils.center_print(resp['detail'], constants.DISPLAY_TERMINAL_COLOR_ERROR)
    exit(0)
        
        
def change_task_priority(task_id: int, new_priority: int):
    mode = auth_utils.get_mode(cfg.config_path)
    if mode == "user":
        resp = requests.patch(
            cfg.base_url + f"/tasks/{task_id}/priority",
            headers=auth_utils.build_jwt_header(cfg.config_path),
            json={
                "new_priority": new_priority
            }
        )
        
        if resp.status_code == 200:
            task = general_utils.bytes2dict(resp.content)
            task_name = task['name']
            display_utils.center_print(f"Successfully changed [{task_name}]'s priority to {new_priority}.", type="success")
            return
        elif resp.status_code == 400:
            err_msg = general_utils.bytes2dict(resp.content)['detail']
            display_utils.center_print(err_msg, type="error")
        else:
            display_utils.center_print(resp.content, type="error")
    elif mode == "guest":
        resp = task_router.change_task_priority(task_id, {"new_priority": new_priority})
        if resp['success']:
            task = resp['updated_task']
            task_name = task['name']
            display_utils.center_print(f"Successfully changed [{task_name}]'s priority to {new_priority}.", type="success")
            return
        else:
            display_utils.center_print(resp['detail'], type="error")
    exit(0)
            

        
def change_task_name(task_id: int, new_name: str):
    mode = auth_utils.get_mode(cfg.config_path)
    if mode == "user":
        resp = requests.patch(
            cfg.base_url + f"/tasks/{task_id}/name",
            headers=auth_utils.build_jwt_header(cfg.config_path),
            json={
                "new_name": new_name
            }
        )
        
        if resp.status_code == 200:
            task = general_utils.bytes2dict(resp.content)
            task_name = task['name']
            display_utils.center_print(f"Successfully changed [{task_name}]'s name to {new_name}.", type="success")
            return
        elif resp.status_code == 400:
            err_msg = general_utils.bytes2dict(resp.content)['detail']
            display_utils.center_print(err_msg, constants.DISPLAY_TERMINAL_COLOR_ERROR)
        else:
            display_utils.center_print(resp.content, constants.DISPLAY_TERMINAL_COLOR_ERROR)
    elif mode == "guest":
        resp = task_router.change_task_name(task_id, {"new_name": new_name})
        if resp['success']:
            task = resp['updated_task']
            task_name = task['name']
            display_utils.center_print(f"Successfully changed [{task_name}]'s name to {new_name}.", type="success")
            return
        else:
            display_utils.center_print(resp['detail'], constants.DISPLAY_TERMINAL_COLOR_ERROR)
    exit(0)
        

def change_task_date(task_id: int, new_date: str):
    mode = auth_utils.get_mode(cfg.config_path)
    if mode == "user":
        resp = requests.patch(
            cfg.base_url + f"/tasks/{task_id}/date",
            headers=auth_utils.build_jwt_header(cfg.config_path),
            json={
                "new_date": new_date
            }
        )
        
        if resp.status_code == 200:
            task = general_utils.bytes2dict(resp.content)
            task_name = task['name']
            display_utils.center_print(f"Successfully changed [{task_name}]'s date to {new_date.split()[0]}.", type="success")
        elif resp.status_code == 400:
            err_msg = general_utils.bytes2dict(resp.content)['detail']
            display_utils.center_print(err_msg, type="error")
        else:
            display_utils.center_print(resp.content, type="error")
    elif mode == "guest":
        resp = task_router.change_task_date(task_id, {"new_date": new_date})
        if resp['success']:
            task = resp['updated_task']
            task_name = task['name']
            display_utils.center_print(f"Successfully changed [{task_name}]'s date to {new_date.split()[0]}.", type="success")
        else:
            display_utils.center_print(resp['detail'], type="error")
    exit(0)


from sqlalchemy import Column, String, Integer, ForeignKey
from sqlalchemy.sql.sqltypes import TIMESTAMP, Boolean, DateTime
from sqlalchemy.sql.expression import text
from sqlalchemy.orm import relationship

from girok.server.src.database import Base


class Task(Base):
    __tablename__ = "task"
    
    task_id = Column(Integer, primary_key=True, autoincrement=True)
    task_category_id = Column(Integer, ForeignKey('task_category.task_category_id', ondelete="CASCADE"), nullable=True)
    name = Column(String(300), nullable=False)
    deadline = Column(DateTime, nullable=False)
    everyday = Column(Boolean, nullable=True)
    weekly_repeat = Column(Integer, nullable=True)
    is_time = Column(Boolean, nullable=False)
    all_day = Column(Boolean, nullable=True)
    priority = Column(Integer, nullable=True)
    color = Column(String(20), nullable=True)
    tag = Column(String(20), nullable=True)
    # created_at = Column(TIMESTAMP(timezone=True), nullable=False, server_default=text('now()'))
